import pygame
import time
import math
import random
import os
from pygame.locals import *
from math import *

scr_width=640
scr_height=480

pygame.mixer.pre_init(44100, 16, 2, 4096)
pygame.font.init()
pygame.init()
pygame.display.set_caption("Tiny Boxes")
#pygame.display.set_icon(pygame.image.load("tinyboxes.ico"))
screen=pygame.display.set_mode((scr_width,scr_height))

lastPlay=0
font=pygame.font.Font("bitType.ttf", 16)
fontBig=pygame.font.Font("bitType.ttf", 32)

sprites=[]
def getSprites():
    return sprites
def clearSprites():
    del sprites[:]

#20x15 grid.
TILE_SIZE=16
tileSet=pygame.image.load("image.png").convert_alpha()

def getTileSubSurf(x,y):
    return tileSet.subsurface(Rect(TILE_SIZE*x,TILE_SIZE*y,TILE_SIZE,TILE_SIZE))

image={}
image['box']=getTileSubSurf(0,0)
image['button']=getTileSubSurf(1,0)
image['player']=getTileSubSurf(2,0)
image['exit']=getTileSubSurf(3,0)
image['wall']=getTileSubSurf(0,1)
image['tele']=getTileSubSurf(1,1)
image['floor']=getTileSubSurf(2,1)
image['ice']=getTileSubSurf(3,1)
image['projectile']=getTileSubSurf(4,0)
image['pointer']=getTileSubSurf(0,3)

#door
for i in xrange(6):
    image['door'+str(i)]=getTileSubSurf(i,4)

tileMap={}
tileMap['box']=(204,102,0)
tileMap['button']=(255,0,0)
tileMap['player']=(255,216,0)
tileMap['wall']=(128,128,128)
tileMap['exit']=(0,0,0)
tileMap['vdoor']=(0,0,255)
tileMap['hdoor']=(0,255,0)
tileMap['tele']=(255,0,255)
tileMap['ice']=(0,255,255)

sound={}
sound['step'] =pygame.mixer.Sound("step.wav")
sound['slide']=pygame.mixer.Sound("slide.wav")
sound['bump'] =pygame.mixer.Sound("bump.wav")
sound['open'] =pygame.mixer.Sound("open.wav")
sound['close']=pygame.mixer.Sound("close.wav")
sound['tele'] =pygame.mixer.Sound("tele.wav")
sound['restart'] =pygame.mixer.Sound("restart.wav")
sound['shoot'] =pygame.mixer.Sound("shoot.wav")
sound['win'] =pygame.mixer.Sound("win.wav")
sound['select'] =pygame.mixer.Sound("select.wav")
sound['menu'] =pygame.mixer.Sound("menu.wav")
sound['music'] =pygame.mixer.Sound("TinyBoxes.wav")

def playSoundOnce(sound):
    global lastPlay, time
    if lastPlay+10<time:
        sound.play()
        lastPlay=time

def wrapRect(rect):
    rect.left=rect.left%320
    rect.top=rect.top%240
    return rect

class Sprite(object):
    def __init__(self,layer=0):
        self.layer=layer
        getSprites().append(self)
        getSprites().sort(key=lambda sprite: sprite.layer)
        self.rect=Rect(0,0,0,0)
        self.image = image['box']
        self.x=0
        self.y=0
        self.sx=0
        self.sy=0
        self.lastSlide=0
        self.teleporting=False
        self.tx=0
        self.ty=0
        #in radians
        self.rot=0

    def __repr__(self):
        return self.getType()+"="+"Layer:"+str(self.layer)

    def render(self,surf):
        if self.rot==0:
            surf.blit(self.image,(self.x,self.y))
        else:
            surf.blit(pygame.transform.rotate(self.image,self.rot),(self.x,self.y))

    def getRect(self):
        return Rect(self.x,self.y,TILE_SIZE,TILE_SIZE)

    def update(self,time):
        if self.teleporting:
            self.rot=degrees(atan2(self.ty-self.y,self.tx-self.x))
            self.move(cos(radians(self.rot))*12,sin(radians(self.rot))*12)
            if self.x<self.tx+8 and self.x>self.tx-8:
                if self.y<self.ty+8 and self.y>self.ty-8:
                    self.teleporting=False
                    self.x=self.tx
                    self.y=self.ty
                    self.rot=0
            
        if self.lastSlide+200<time:
            self.lastSlide=time
            if self.isSliding():
                self.tryMove(self.sx,self.sy)

    def setPos(self,x,y):
        self.x=x%320
        self.y=y%240

    def move(self,dx,dy):
        self.setPos(self.x+dx,self.y+dy)

    def isSliding(self):
        return (self.sx!=0)|(self.sy!=0)

    def tryMove(self,dx,dy,canTele=True):
        self.sx=0
        self.sy=0

        btnOff=None
        #will we be moving off of a button?
        obj=collides(self.getRect(), [self])
        if obj!=None:
            if obj.getType()=="Button":
                btnOff=obj
        self.setPos(self.x+dx,self.y+dy)

        if collides(self.getRect(), [self]):
            for obj in collidesAll(self.getRect(), self):
                objType=obj.getType()
                if objType=="Button":
                    obj.setPressed(True)
                    continue
                elif objType=="Door":
                    if obj.open:
                        continue
                elif objType=="Tele":
                    #is it open?
                    tele=otherTele(obj.mapID, obj)
                    if tele ==None:
                        continue
                    if collides(tele.getRect(), [tele]):
                        continue
                    else:
                        #is this one clear?
                        if not collides(obj.getRect(),[obj,self]):
                            #can we teleport?
                            if canTele:
                                self.tx=tele.x
                                self.ty=tele.y
                                self.teleporting=True
                                playSoundOnce(sound['tele'])
                                continue
                            else:
                                continue
                elif objType=="Ice":
                    #Set the slide
                    self.sy=dy
                    self.sx=dx
                    continue
                #Undo the move
                self.setPos(self.x-dx,self.y-dy)
                self.sx=0
                self.sy=0
                return False
        if btnOff!=None:
            btnOff.setPressed(False)
            print 'button off'
        return True

    def getType(self):
        return "Sprite"

class Box(Sprite):
    def __init__(self,x,y):
        super(Box, self).__init__(50)
        self.image = image['box']
        self.x=x
        self.y=y

    def getType(self):
        return "Box"

class Wall(Sprite):
    def __init__(self,x,y):
        super(Wall, self).__init__(0)
        self.image = image['wall']
        self.x=x
        self.y=y

    def getType(self):
        return "Wall"

class Exit(Sprite):
    def __init__(self,x,y):
        super(Exit, self).__init__(-100)
        self.image = image['exit']
        self.x=x
        self.y=y

    def getType(self):
        return "Exit"


class Button(Sprite):
    def __init__(self,x,y):
        super(Button, self).__init__(-100)
        self.image = image['button']
        self.x=x
        self.y=y
        self.pressed=False
        self.mapID=-1

    def getType(self):
        return "Button"

    def render(self,surf):
        surf.blit(self.image,(self.x,self.y))

    def setPressed(self, pressed):
        self.pressed=pressed
        updateDoorChannel(self.mapID)

class Tele(Sprite):
    def __init__(self,x,y):
        super(Tele, self).__init__(-100)
        self.image = image['tele']
        self.x=x
        self.y=y
        self.pressed=False
        self.mapID=-1

    def getType(self):
        return "Tele"

class Ice(Sprite):
    def __init__(self,x,y):
        super(Ice, self).__init__(-100)
        self.image = image['ice']
        self.x=x
        self.y=y

    def getType(self):
        return "Ice"

class Door(Sprite):
    def __init__(self,x,y):
        super(Door, self).__init__(100)
        self.image = image['door0']
        self.x=x
        self.y=y
        self.open=False
        self.channelOn=False
        self.frame=0
        self.frames=10
        self.mapID=-1

    def getType(self):
        return "Door"

    def render(self,surf):
        surf.blit(pygame.transform.rotate(self.image,self.rot),(self.x,self.y))
        if self.open==True:
            #is it closing?
            if self.frame!=self.frames:
                self.frame+=1
                self.image=image['door'+str(int(self.frame/2))]
        else:
            #is it opening?
            if self.frame!=0:
                self.frame-=1
                self.image=image['door'+str(int(self.frame/2))]

    def setOpen(self, open_):
        self.channelOn=open_
        if collides(self.getRect(), [self]):
            return
        if self.open==open_:
            return
        self.open=open_
        if open_:
            playSoundOnce(sound['open'])
        else:
            playSoundOnce(sound['close'])

    def update(self,time):
        if (not collides(self.getRect(), [self]))and self.channelOn==False:
            self.open=False

    def setAngle(self,angle):
        self.rot=angle%360

    def __repr__(self):
        return self.getType()+"="+"Layer:"+str(self.layer)+" Open:"+str(self.open)

class Player(Sprite):
    def __init__(self):
        super(Player, self).__init__(50)
        self.image = image['player']
        self.x=0
        self.y=0
        self.fx=0
        self.fy=0
        self.flying=False
        self.teleporting=False
          
    def controllable(self):
        return ((not self.flying) and (not self.teleporting))

    def render(self,surf):
        if not self.flying:
            surf.blit(pygame.transform.rotate(self.image,self.rot),(self.x,self.y))
        else:
            surf.blit(pygame.transform.rotate(image['projectile'],self.rot),(self.x,self.y))

    def setAngle(self,angle):
        if not self.flying:
            self.rot=angle%360

    def rotate(self, amt):
        self.setAngle(self.rot+amt)

    def getType(self):
        return "Player"

    def update(self,time):
        if self.flying:
            self.flying=self.tryMove(self.fx,self.fy)
        else:
            super(Player, self).update(time)

    def interactiveMove(self, x,y):
        global countTime
        countTime=True
        if self.controllable(): #Can't move while flying!
            #Can't change direction while sliding
            if self.isSliding():
                sound['bump'].play()
                return
            dx=TILE_SIZE*x
            dy=TILE_SIZE*y
            if self.tryMove(dx,dy):
                sound['step'].play()
            else:
                #see if there is a box to move
                for obj in collidesAll(wrapRect(self.getRect().move(dx,dy)),self):
                    if obj.getType()=="Exit":
                        #Next Level!
                        self.move(dx,dy)
                        endLevel()
                    elif obj==None:
                        print "hackz"
                        return
                    elif obj.getType()=="Box":
                        if obj.tryMove(dx,dy):
                            self.tryMove(dx,dy,False)
                            sound['slide'].play()
                            return
                sound['bump'].play()
        else:
            sound['bump'].play()
            return
                
    def shoot(self):
        global countTime
        countTime=True
        if self.controllable():
            sound['shoot'].play()
            self.flying=True
            self.fx=cos(radians(self.rot))*TILE_SIZE
            self.fy=-sin(radians(self.rot))*TILE_SIZE
        else:
            sound['bump'].play()

def invert(img):
   inv = pygame.Surface(img.get_rect().size, pygame.SRCALPHA)
   inv.fill((255,255,255,255))
   inv.blit(img, (0,0), None, BLEND_RGB_SUB)
   return inv

def updateDoorChannel(chan):
    value=False
    for i in xrange(len(sprites)):
        if sprites[i].getType()=="Button":
            if sprites[i].mapID==chan:
                if sprites[i].pressed:
                    value=True
    #print "Updating channel:"+str(channel)+":"+str(value)
    for i in xrange(len(sprites)):
        if sprites[i].getType()=="Door":
            if sprites[i].mapID==chan:
                sprites[i].setOpen(value)

def otherTele(channel, exclude):
    #print "Teleporting channel:"+str(channel)+":"+str(value)
    for i in xrange(len(sprites)):
        if sprites[i].getType()=="Tele":
            if sprites[i]!=exclude:
                if sprites[i].mapID==channel:
                    return sprites[i]

def collides(rect, exclude):
    for i in xrange(len(sprites)):
        if exclude.count(sprites[i])>0:
            continue
        if rect.colliderect(sprites[i].getRect()):
            return sprites[i]

def collidesAll(rect, exclude):
    collisions=[]
    for i in xrange(len(sprites)):
        if sprites[i]==exclude:
            continue
        if rect.colliderect(sprites[i].getRect()):
            collisions.append(sprites[i])
    return collisions

#Global Player defn

def loadLevel(fname):
    print "Loading From:" +fname
    resetTime()  
    clearSprites()
    global player
    player=Player()
    level=pygame.image.load(fname)
    level.convert_alpha()
    level.lock()

    iceLevel=False
    c=level.get_at((0,0))
    color=(c.r,c.g,c.b)
    if color==tileMap['ice']:
        iceLevel=True
        for x in xrange(level.get_width()):
            for y in xrange(level.get_height()):
                Ice(x*TILE_SIZE,y*TILE_SIZE)

    for x in xrange(level.get_width()):
        for y in xrange(level.get_height()):
            c=level.get_at((x,y))
            data=c.a
            color=(c.r,c.g,c.b)
            if color==tileMap['box']:
                Box(x*TILE_SIZE,y*TILE_SIZE)
            elif color==tileMap['button']:
                Button(x*TILE_SIZE,y*TILE_SIZE).mapID=data
            elif color==tileMap['vdoor']:
                Door(x*TILE_SIZE,y*TILE_SIZE).mapID=data
            elif color==tileMap['hdoor']:
                door=Door(x*TILE_SIZE,y*TILE_SIZE)
                door.mapID=data
                door.setAngle(270)
            elif color==tileMap['player']:
                player.x=x*TILE_SIZE
                player.y=y*TILE_SIZE
            elif color==tileMap['wall']:
                Wall(x*TILE_SIZE,y*TILE_SIZE)
            elif color==tileMap['tele']:
                Tele(x*TILE_SIZE,y*TILE_SIZE).mapID=data
            elif color==tileMap['ice']:
                if not iceLevel:
                    Ice(x*TILE_SIZE,y*TILE_SIZE)
            elif color==tileMap['exit']:
                Exit(x*TILE_SIZE,y*TILE_SIZE)
    level.unlock()
    time=0
    return

#init levels
levels=[]
levelDir="levels/"
for file in os.listdir(levelDir):
    dirfile = os.path.join(levelDir, file)
    if os.path.isfile(dirfile):#Check aginst directories
        if dirfile.endswith(".png"):
            levels.append(dirfile)
    levels.sort()

level=0

def toggleFullscreen():
    global fullscreen, screenSize
    if fullscreen:
        pygame.display.set_mode((scr_width,scr_height),0,32)
        fullscreen=False
    else:
        #set to max res, then scale in draw operation
        pygame.display.set_mode(screenSize,FULLSCREEN,32)
        fullscreen=True

def updateScreen():
    global fullscreen,size,background,offset
    if fullscreen:
        screen.blit(pygame.transform.scale(background,size),(offset,0))
    else:
        screen.blit(background, (0,0))
    pygame.display.flip()

def endLevel():
    sound['win'].play()
    global state
    state=TRANSITION

def nextLevel():
    global level,state
    level+=1    
    if level==len(levels):
        print "End of game!"
        level=0
        state=END
    else:
        reloadLevel()

def reloadLevel():
    global state
    state=GAME
    loadLevel(levels[level])

def renderAll():
    global pxlBackground, time,fullscreen,size
    for x in xrange(20):
        for y in xrange(15):
            pxlBackground.blit(image['floor'],(x*TILE_SIZE,y*TILE_SIZE))

    for i in xrange(len(sprites)):
        sprites[i].render(pxlBackground)

    text=font.render(format(time/1000.0,"7.3f")+"s", False,  (255,255,255))

    pxlBackground.blit(text,(2,0))
    pygame.transform.scale(pxlBackground, (scr_width, scr_height),background)
    updateScreen()

#initilize Everything
time=0
lastUpdate=time

def resetTime():
    global time, lastUpdate, countTime
    time=0
    lastUpdate=0
    countTime=False
    
background=pygame.Surface((scr_width, scr_height), HWSURFACE, depth=32)
pxlBackground=pygame.Surface((scr_width/2, scr_height/2), HWSURFACE, depth=32)

GAME=1
MENU=0
END=2
HELP=3
TRANSITION=4
state=MENU

countTime=False

resetTime()

screenSize=pygame.display.list_modes()[0]
fullscreen=False
size=(int(screenSize[1]*(float(scr_width)/scr_height)),screenSize[1])
offset=(screenSize[0]-size[0])/2

def run():
    title=pygame.image.load("title.png").convert()
    help=pygame.image.load("help.png").convert()
    end=pygame.image.load("end.png").convert()
    next=pygame.image.load("next.png").convert()
    replay=pygame.image.load("replay.png").convert()
    select=pygame.image.load("select.png").convert_alpha()
    transBack=pygame.image.load("transback.png").convert()
    sound['music'].play(-1)#loop our background music
    sound['music'].set_volume(.8)
    pointer=0

    global time,state,lastUpdate

    clock = pygame.time.Clock()
    fps=60
    selection=0
    screen = pygame.display.set_mode((scr_width, scr_height))
    run=True

    while run:
        if state==END:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run=False
                if event.type == KEYDOWN:
                    if event.key == K_F11:
                        toggleFullscreen()
                    else:
                        state=MENU
            pxlBackground.blit(end,(0,0))

            pygame.transform.scale(pxlBackground, (scr_width, scr_height),background)

            updateScreen()

            clock.tick(fps)
        if state==TRANSITION:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run=False
                if event.type == KEYDOWN:
                    if event.key==K_a:
                        if selection>0:
                            selection-=1
                            sound['select'].play()
                    elif event.key==K_d:
                        if selection<1:
                            selection+=1
                            sound['select'].play()
                    elif event.key == K_F11:
                        toggleFullscreen()
                    elif event.key==K_SPACE:
                        if selection==0:
                            reloadLevel()
                        else:
                            nextLevel()
            pxlBackground.blit(transBack,(0,0))

            text=fontBig.render(format(time/1000.0,"7.3f")+"s", False,  (255,255,255))

            pxlBackground.blit(text, (97,89))

            pxlBackground.blit(replay, (80,146))
            pxlBackground.blit(next, (176,146))
            pxlBackground.blit(select, (selection*96+80,146))
    

            pygame.transform.scale(pxlBackground, (scr_width, scr_height),background)
            updateScreen()
            clock.tick(fps)
            
        elif state==HELP:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run=False
                if event.type == KEYDOWN:
                    if event.key == K_F11:
                        toggleFullscreen()
                    else:
                        sound['menu'].play()
                        state=MENU
            pxlBackground.blit(help,(0,0))
            pygame.transform.scale(pxlBackground, (scr_width, scr_height),background)
            updateScreen()

            clock.tick(fps)
        if state==MENU:
            for event in pygame.event.get():
                 if event.type == pygame.QUIT:
                    run=False
                 if event.type == KEYDOWN:
                    if event.key==K_w:
                        if pointer>0:
                            pointer-=1
                            sound['select'].play()
                    elif event.key==K_s:
                        if pointer<=1:
                            pointer+=1
                            sound['select'].play()
                    elif event.key==K_SPACE:
                        sound['menu'].play()
                        if pointer==0:
                            state=GAME
                            level=0
                            time=0
                            reloadLevel()
                        elif pointer==1:
                            state=HELP
                        elif pointer==2:
                            run=False
                    elif event.key == K_F11:
                        toggleFullscreen()

            pxlBackground.blit(title,(0,0))
            pxlBackground.blit(image['pointer'],(182,89+30*pointer))

            pygame.transform.scale(pxlBackground, (scr_width, scr_height),background)
            updateScreen()

            clock.tick(fps)
        elif state==GAME:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    run=False
                if event.type == KEYDOWN:
                    if event.key == K_F11:
                        toggleFullscreen()
                    elif event.key==K_w:
                        player.setAngle(90)
                        player.interactiveMove(0,-1)
                    elif event.key==K_ESCAPE:
                        state=MENU
                        level=0
                    elif event.key==K_s:
                        player.setAngle(270)
                        player.interactiveMove(0,1)
                    elif event.key==K_a:
                        player.setAngle(180)
                        player.interactiveMove(-1,0)
                    elif event.key==K_d:
                        player.setAngle(0)
                        player.interactiveMove(1,0)
                    elif event.key==K_SPACE:
                        player.shoot()
                    elif event.key==K_r:
                        reloadLevel()
                        playSoundOnce(sound['restart'])

            if time>lastUpdate*50:
                for i in xrange(len(sprites)):
                    sprites[i].update(time)
                lastUpdate+=1
                #print str(time)

            renderAll()

            dtime=clock.tick(fps)

            if countTime:
                time+=dtime
    sound['music'].fadeout(1500)
    pygame.display.quit()
    #exit(0)
    print "===QUIT==="

if __name__ == "__main__":
    run()
